from __future__ import annotations

import numpy as np
import pytest
from xarray import DataArray, Dataset

from openlifu.plan.solution_analysis import get_offset_grid


@pytest.fixture()
def example_xarr() -> DataArray:
    rng = np.random.default_rng(147)
    return Dataset(
            {
                'p': DataArray(
                    data=rng.random((3, 2, 3)),
                    dims=["x", "y", "z"],
                    attrs={'units': "Pa"}
                )
            },
            coords={
                'x': DataArray(dims=["x"], data=np.linspace(0, 1, 3), attrs={'units': "mm"}),
                'y': DataArray(dims=["y"], data=np.linspace(0, 1, 2), attrs={'units': "mm"}),
                'z': DataArray(dims=["z"], data=np.linspace(0, 1, 3), attrs={'units': "mm"})
            }
        )

def test_offset_grid(example_xarr: Dataset):
    """Test that the distance grid from the focus point is correct."""
    expected = np.array([
        [[[ 0. ,  0. , -1. ],
         [ 0. ,  0. , -0.5],
         [ 0. ,  0. ,  0. ]],

        [[ 0. ,  1. , -1. ],
         [ 0. ,  1. , -0.5],
         [ 0. ,  1. ,  0. ]]],


       [[[ 0.5,  0. , -1. ],
         [ 0.5,  0. , -0.5],
         [ 0.5,  0. ,  0. ]],

        [[ 0.5,  1. , -1. ],
         [ 0.5,  1. , -0.5],
         [ 0.5,  1. ,  0. ]]],


       [[[ 1. ,  0. , -1. ],
         [ 1. ,  0. , -0.5],
         [ 1. ,  0. ,  0. ]],

        [[ 1. ,  1. , -1. ],
         [ 1. ,  1. , -0.5],
         [ 1. ,  1. ,  0. ]]]])
    offset = get_offset_grid(example_xarr, [0.0, 0.0, 1.0], as_dataset=False)

    np.testing.assert_almost_equal(offset, expected)



from __future__ import annotations

from openlifu.io.LIFUInterface import LIFUInterface

# set PYTHONPATH=%cd%\src;%PYTHONPATH%
# python notebooks/test_multiple_modules.py
"""
Test script to automate:
1. Connect to the device.
2. Test HVController: Turn HV on/off and check voltage.
3. Test Device functionality.
"""
print("Starting LIFU Test Script...")
interface = LIFUInterface()
tx_connected, hv_connected = interface.is_device_connected()
if tx_connected and hv_connected:
    print("LIFU Device Fully connected.")
else:
    print(f'LIFU Device NOT Fully Connected. TX: {tx_connected}, HV: {hv_connected}')

print("Ping the device")
interface.txdevice.ping()

print("Enumerate TX7332 chips")
num_tx_devices = interface.txdevice.enum_tx7332_devices()
if num_tx_devices > 0:
    print(f"Number of TX7332 devices found: {num_tx_devices}")

    print("Write Demo Registers to TX7332 chips")
    for device_index in range(num_tx_devices):
        interface.txdevice.demo_tx7332(device_index)

    print("Starting Trigger...")
    if interface.txdevice.start_trigger():
        print("Trigger Running Press enter to STOP:")
        input()  # Wait for the user to press Enter
        if interface.txdevice.stop_trigger():
            print("Trigger stopped successfully.")
        else:
            print("Failed to stop trigger.")
    else:
        print("Failed to start trigger.")

else:
    raise Exception("No TX7332 devices found.")



from __future__ import annotations

import sys

import base58

from openlifu.io.LIFUInterface import LIFUInterface

# set PYTHONPATH=%cd%\src;%PYTHONPATH%
# python notebooks/test_console.py
"""
Test script to automate:
1. Connect to the device.
2. Test HVController: Turn HV on/off and check voltage.
3. Test Device functionality.
"""
print("Starting LIFU Test Script...")
interface = LIFUInterface(TX_test_mode=False)
tx_connected, hv_connected = interface.is_device_connected()
if tx_connected and hv_connected:
    print("LIFU Device Fully connected.")
else:
    print(f'LIFU Device NOT Fully Connected. TX: {tx_connected}, HV: {hv_connected}')

if not hv_connected:
       print("HV Controller not connected.")
       sys.exit()

print("Ping the device")
interface.hvcontroller.ping()

print("Toggle LED")
interface.hvcontroller.toggle_led()

print("Get Version")
version = interface.hvcontroller.get_version()
print(f"Version: {version}")

print("Echo Data")
echo, echo_len = interface.hvcontroller.echo(echo_data=b'Hello LIFU!')
if echo_len > 0:
    print(f"Echo: {echo.decode('utf-8')}")  # Echo: Hello LIFU!
else:
    print("Echo failed.")

print("Get HW ID")
hw_id = interface.hvcontroller.get_hardware_id()
print(f"HW ID: {hw_id}")
encoded_id = base58.b58encode(bytes.fromhex(hw_id)).decode()
print(f"OW-LIFU-CON-{encoded_id}")

print("Get Temperature1")
temp1 = interface.hvcontroller.get_temperature1()
print(f"Temperature1: {temp1}")

print("Get Temperature2")
temp2 = interface.hvcontroller.get_temperature2()
print(f"Temperature2: {temp2}")

print("Set Bottom Fan Speed to 20%")
btfan_speed = interface.hvcontroller.set_fan_speed(fan_id=0, fan_speed=20)
print(f"Bottom Fan Speed: {btfan_speed}")

print("Set Top Fan Speed to 40%")
tpfan_speed = interface.hvcontroller.set_fan_speed(fan_id=1, fan_speed=40)
print(f"Bottom Fan Speed: {tpfan_speed}")

print("Get Bottom Fan Speed")
btfan_speed = interface.hvcontroller.get_fan_speed(fan_id=0)
print(f"Bottom Fan Speed: {btfan_speed}")

print("Get Top Fan Speed")
tpfan_speed = interface.hvcontroller.get_fan_speed(fan_id=1)
print(f"Bottom Fan Speed: {tpfan_speed}")

print("Set RGB LED")
rgb_led = interface.hvcontroller.set_rgb_led(rgb_state=2)
print(f"RGB STATE: {rgb_led}")

print("Get RGB LED")
rgb_led_state = interface.hvcontroller.get_rgb_led()
print(f"RGB STATE: {rgb_led_state}")

print("Test 12V...")
if interface.hvcontroller.turn_12v_on():
    print("12V ON Press enter to TURN OFF:")
    input()  # Wait for the user to press Enter
    if interface.hvcontroller.turn_12v_off():
        print("12V OFF.")
    else:
        print("Failed to turn off 12V")
else:
    print("Failed to turn on 12V.")

# Set High Voltage Level
print("Set HV Power to +/- 85V")
if interface.hvcontroller.set_voltage(voltage=75.0):
    print("Voltage set to 85.0 V.")
else:
    print("Failed to set voltage.")

# Get Set High Voltage Setting
print("Get Current HV Voltage")
read_voltage = interface.hvcontroller.get_voltage()
print(f"HV Voltage {read_voltage} V.")


print("Test HV Supply...")
if interface.hvcontroller.turn_hv_on():
# Get Set High Voltage Setting
    read_voltage = interface.hvcontroller.get_voltage()
    print(f"HV Voltage {read_voltage} V.")
    print("HV ON Press enter to TURN OFF:")
    input()  # Wait for the user to press Enter
    if interface.hvcontroller.turn_hv_off():
        print("HV OFF.")
    else:
        print("Failed to turn off HV")
else:
    print("Failed to turn on HV.")

print("Reset DevConsoleice:")
# Ask the user for confirmation
user_input = input("Do you want to reset the Console? (y/n): ").strip().lower()

if user_input == 'y':
    if interface.hvcontroller.soft_reset():
        print("Reset Successful.")
elif user_input == 'n':
    print("Reset canceled.")
else:
    print("Invalid input. Please enter 'y' or 'n'.")



from __future__ import annotations

from .apodmethod import ApodizationMethod
from .maxangle import MaxAngle
from .piecewiselinear import PiecewiseLinear
from .uniform import Uniform

__all__ = [
    "ApodizationMethod",
    "Uniform",
    "MaxAngle",
    "PiecewiseLinear",
]



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Annotated, Any, Dict, List, Sequence, Tuple

import numpy as np

from openlifu.geo import (
    cartesian_to_spherical_vectorized,
    spherical_coordinate_basis,
    spherical_to_cartesian,
    spherical_to_cartesian_vectorized,
)
from openlifu.seg.skinseg import (
    compute_foreground_mask,
    create_closed_surface_from_labelmap,
    spherical_interpolator_from_mesh,
    vtk_img_from_array_and_affine,
)
from openlifu.util.annotations import OpenLIFUFieldData
from openlifu.util.dict_conversion import DictMixin
from openlifu.util.units import getunitconversion

log = logging.getLogger("VirtualFit")

ras2asl_3x3 = np.array([[0,1,0],[0,0,1],[-1,0,0]], dtype=float) # ASL means Anterior-Superior-Left coordinates
asl2ras_3x3 = ras2asl_3x3.transpose()

@dataclass
class VirtualFitOptions(DictMixin):
    """Parameters to configure the `virtual_fit` algorithm.

    The terms 'pitch' and 'yaw' used here refer to the following target-centric angular coordinates in patient space:
        pitch: The angle between the anterior axis through the target and the ray from the target to the projection of
            a given point into the anterior-superior plane.
        yaw: The angle between the anterior-superior plane through the target and the ray from the target to a given point.

    Another way to describe them in terms of standard spherical coordinates centered at the target in ASL (anterior-superior-left) space:
        pitch: The azimuthal spherical coordinate.
        yaw: 90 degrees minus the polar spherical coordinate.
    """

    units: Annotated[str, OpenLIFUFieldData("Length units", "The units of length used in the length attributes of this class")] = "mm"
    """The units of length used in the length attributes of this class"""

    transducer_steering_center_distance: Annotated[float, OpenLIFUFieldData("Steering center distance", "Distance from the transducer origin axially to the center of the steering zone in the units `units`")] = 50.
    """Distance from the transducer origin axially to the center of the steering zone in the units `units`"""

    steering_limits: Annotated[Tuple[Tuple[float, float], Tuple[float, float], Tuple[float, float]],
                               OpenLIFUFieldData("Steering limits", "Steering bounds along each axis from the transducer origin, in the units `units`")] = ((-50, 50), (-50, 50), (-50, 50))
    """Distance from the transducer origin axially to the center of the steering zone in the units `units`"""

    pitch_range: Annotated[Tuple[float, float], OpenLIFUFieldData("Pitch range (deg)", "Range of pitches to include in the transducer fitting search grid, in degrees")] = (-10, 150)
    """Range of pitches to include in the transducer fitting search grid, in degrees"""

    pitch_step: Annotated[float, OpenLIFUFieldData("Pitch step size (deg)", "Pitch step size when forming the transducer fitting search grid, in degrees")] = 5
    """Pitch step size when forming the transducer fitting search grid, in degrees"""

    yaw_range: Annotated[Tuple[float, float], OpenLIFUFieldData("Yaw range (deg)", "Range of yaws to include in the transducer fitting search grid, in degrees")] = (-65, 65)
    """Range of yaws to include in the transducer fitting search grid, in degrees"""

    yaw_step: Annotated[float, OpenLIFUFieldData("Yaw step size (deg)", "Yaw step size when forming the transducer fitting search grid, in degrees")] = 5
    """Yaw step size when forming the transducer fitting search grid, in degrees"""

    planefit_dyaw_extent: Annotated[float, OpenLIFUFieldData("Plane fit yaw extent", "Left and right extents of the point grid to be used for plane fitting along the local yaw axes, in units of `units`")] = 15
    """Left and right extents of the point grid to be used for plane fitting along the local yaw axes,
    in units of `units`. The plane fitting point grid will be twice this size, since this is left
    and right extents. (Note that this has units of length, not angle!)"""

    planefit_dyaw_step: Annotated[float, OpenLIFUFieldData("Plane fit yaw step", "Local yaw axis step size to use when constructing plane fitting grids. In spatial units of `units`")] = 3
    """Local yaw axis step size to use when constructing plane fitting grids. In spatial units of `units`."""

    planefit_dpitch_extent: Annotated[float, OpenLIFUFieldData("Plane fit pitch extent", "Left and right extents of the point grid to be used for plane fitting along the local pitch axes, in spatial units of `units`")] = 15
    """Left and right extents of the point grid to be used for plane fitting along the local pitch axes,
    in spatial units of `units`. The plane fitting point grid will be twice this size, since this is left
    and right extents."""

    planefit_dpitch_step: Annotated[float, OpenLIFUFieldData("Plane fit pitch step", "Local pitch axis step size to use when constructing plane fitting grids. In spatial units of `units`")] = 3
    """Local pitch axis step size to use when constructing plane fitting grids. In spatial units of `units`."""

    def to_units(self, target_units: str) -> VirtualFitOptions:
        """Do unit conversion and return a version of this VirtualFitOptions that uses
        `target_units` as the units for all attributes that have units of length."""
        conversion_factor = getunitconversion(from_unit = self.units, to_unit=target_units)
        return VirtualFitOptions(
            units = target_units,
            transducer_steering_center_distance = conversion_factor * self.transducer_steering_center_distance,
            steering_limits = tuple(map(tuple,conversion_factor*np.array(self.steering_limits))),
            pitch_range = self.pitch_range,
            pitch_step = self.pitch_step,
            yaw_range = self.yaw_range,
            yaw_step = self.yaw_step,
            planefit_dyaw_extent = conversion_factor * self.planefit_dyaw_extent,
            planefit_dyaw_step = conversion_factor * self.planefit_dyaw_step,
            planefit_dpitch_extent = conversion_factor * self.planefit_dpitch_extent,
            planefit_dpitch_step = conversion_factor * self.planefit_dpitch_step,
        )

    @staticmethod
    def from_dict(parameter_dict: Dict[str,Any]) -> VirtualFitOptions: # Override DictMixin here
        parameter_dict["pitch_range"] = tuple(parameter_dict["pitch_range"])
        parameter_dict["yaw_range"] = tuple(parameter_dict["yaw_range"])
        parameter_dict["steering_limits"] = tuple(map(tuple,parameter_dict["steering_limits"]))
        return VirtualFitOptions(**parameter_dict)


def virtual_fit(
    volume_array : np.ndarray,
    volume_affine_RAS : np.ndarray,
    units: str,
    target_RAS : Sequence[float],
    standoff_transform : np.ndarray,
    options : VirtualFitOptions,
) -> List[np.ndarray]:
    """Run patient-specific "virtual fitting" algorithm, suggesting a series of candidate transducer
    transforms for optimal sonicaiton of a given target.

    Args:
        volume_array: A 3D volume MRI
        volume_affine_RAS: A 4x4 affine transform that maps `volume_array` into RAS space with certain units
        units: The spatial units of the RAS space into which volume_affine_RAS maps
        target_RAS: A 3D point, in the coordinates and units of `volume_affine_RAS` (the `units` argument)
        standoff_transform: See the documentation of `create_standoff_transform` or
            `Transducer.standoff_transform` for the meaning of this. Here it should be provided in the
            units `units`. The method `Transducer.get_standoff_transform_in_units` is useful for getting this.
        options : Virtual fitting algorithm configuration. See the `VirtualFitOptions` documentation.

    Returns: A list of transducer transform candidates sorted starting from the best-scoring one. The transforms map transducer space
        into LPS space, and they are in the same units as the RAS space of `volume_affine_RAS` (aka the `units` argument).
    """

    # Express all virtual fit options in the units of volume_affine_RAS, i.e. the physical space of the volume
    options = options.to_units(units)
    pitch_range = options.pitch_range
    pitch_step = options.pitch_step
    yaw_range = options.yaw_range
    yaw_step = options.yaw_step
    transducer_steering_center_distance = options.transducer_steering_center_distance
    steering_limits = options.steering_limits
    planefit_dyaw_extent = options.planefit_dyaw_extent
    planefit_dyaw_step = options.planefit_dyaw_step
    planefit_dpitch_extent = options.planefit_dpitch_extent
    planefit_dpitch_step = options.planefit_dpitch_step

    log.info("Computing foreground mask...")
    foreground_mask_array = compute_foreground_mask(volume_array)
    foreground_mask_vtk_image = vtk_img_from_array_and_affine(foreground_mask_array, volume_affine_RAS)
    log.info("Creating closed surface from labelmap...")
    skin_mesh = create_closed_surface_from_labelmap(foreground_mask_vtk_image)
    log.info("Building skin interpolator...")
    skin_interpolator = spherical_interpolator_from_mesh(
        surface_mesh = skin_mesh,
        origin = target_RAS,
        xyz_direction_columns = asl2ras_3x3, # surface mesh was in RAS, so here spherical coordinates are placed on ASL space
    )

    # Useful transforms to and from the skin_interpolator ASL space and between RAS and LPS
    # Note that ASL is a left-handed coordinate system while RAS and LPS are right-handed.
    interpolator2ras = np.eye(4)
    interpolator2ras[:3,:3] = asl2ras_3x3
    interpolator2ras[:3,3] = target_RAS
    ras_lps_swap = np.diag([-1.,-1,1,1])
    interpolator2lps = ras_lps_swap @ interpolator2ras

    # Useful arrays for vectorized comparisons
    steering_mins = np.array([sl[0] for sl in steering_limits], dtype=float) # shape (3,). It is the lat,ele,ax steering min
    steering_maxs = np.array([sl[1] for sl in steering_limits], dtype=float) # shape (3,). It is the lat,ele,ax steering max

    log.info("Searching through candidate transducer poses...")

    # Construct search grid
    theta_sequence = np.arange(90 - yaw_range[-1], 90 - yaw_range[0], yaw_step)
    phi_sequence = np.arange(pitch_range[0], pitch_range[-1], pitch_step)
    theta_grid, phi_grid = np.meshgrid(theta_sequence, phi_sequence, indexing="ij") # each has shape (number of thetas, number of phis)
    num_thetas, num_phis = theta_grid.shape
    num_search_points = num_thetas*num_phis
    thetas = theta_grid.reshape(num_search_points)
    phis = phi_grid.reshape(num_search_points)

    transducer_poses = np.empty((num_search_points,4,4), dtype=float)
    in_bounds = np.zeros(shape=num_search_points, dtype=bool)
    steering_dists = np.zeros(shape=num_search_points, dtype=float)

    for i in range(num_search_points):
        theta_rad, phi_rad = thetas[i]*np.pi/180, phis[i]*np.pi/180

        # Cartesian coordinate location of the point at which we are fitting a plane
        point = np.array(spherical_to_cartesian(skin_interpolator(theta_rad, phi_rad), theta_rad, phi_rad))

        # Build plane fitting grid in the spherical coordinate basis theta-phi plane, which we will later project back onto the skin surface
        dtheta_sequence = np.arange(-planefit_dyaw_extent, planefit_dyaw_extent + planefit_dyaw_step, planefit_dyaw_step)
        dphi_sequence = np.arange(-planefit_dpitch_extent, planefit_dpitch_extent + planefit_dpitch_step, planefit_dpitch_step)
        dtheta_grid, dphi_grid = np.meshgrid(dtheta_sequence, dphi_sequence, indexing='ij')

        r_hat, theta_hat, phi_hat = spherical_coordinate_basis(theta_rad,phi_rad)
        planefit_points_unprojected_cartesian = (
            point.reshape((1,1,3))
            + dtheta_grid[...,np.newaxis] * theta_hat.reshape(1,1,3) # shape (num dthetas, num dphis, 3)
            + dphi_grid[...,np.newaxis] * phi_hat.reshape(1,1,3) # shape (num dthetas, num dphis, 3)
        ) # shape (num dthetas, num dphis, 3)

        planefit_points_unprojected_spherical = cartesian_to_spherical_vectorized(
            planefit_points_unprojected_cartesian
        ) # shape (num dthetas, num dphis, 3)
        skin_projected_r_values = skin_interpolator(planefit_points_unprojected_spherical[...,1:]) # shape (num dthetas, num dphis) # TODO adjust docstrings to demand a *vectorizable* spherical interpolator
        planefit_points_cartesian = spherical_to_cartesian_vectorized( # Could instead renormalize planefit_points_unprojected_cartesian, not sure if it would give a speedup versus this
            np.stack([
                skin_projected_r_values, # New r values after projection to skin
                planefit_points_unprojected_spherical[...,1], # Same old theta values
                planefit_points_unprojected_spherical[...,2], # Same old phi values
            ], axis=-1)
        )

        # Fit the best plane to these points among the planes that pass through `point`. Here we find the normal vector to the plane.
        plane_normal = np.linalg.svd(
            planefit_points_cartesian.reshape(-1,3)-point.reshape(1,3),
            full_matrices=False, # we don't need the left-singular vectors anyway, so this speeds things up
        ).Vh[-1] # The right-singular vector corresponding to the smallest singular value


        # Transducer axial axis: Parallel to plane_normal, but points towards rather than away from the origin.
        plane_normal_norm = np.linalg.norm(plane_normal)
        if plane_normal_norm < 1e-10:
            continue # Bad geometry at this location, so it's not a virtual fit candidate
        transducer_z = - np.sign(np.dot(plane_normal,point)) * plane_normal / plane_normal_norm

        # Transducer elevational axis: Phi-hat, but then with its component along transducer_z eliminated. This orients the transducer "up" if this were forehead, for example.
        transducer_y = phi_hat - np.dot(phi_hat, transducer_z) * transducer_z
        transducer_y_norm = np.linalg.norm(transducer_y)
        if transducer_y_norm < 1e-10:
            continue # Bad geometry at this location, so it's not a virtual fit candidate
        transducer_y = transducer_y / transducer_y_norm

        # Transducer lateral axis, here simply the only remaining choice to keep it a left handed coordinate system
        # (ASL is left-handed, so the transducer axes must be left-handed to make for an orientation-preserving transducer transform)
        transducer_x = np.cross(transducer_z, transducer_y)

        transducer_transform = np.array(
            [
                [*transducer_x, 0],
                [*transducer_y, 0],
                [*transducer_z, 0],
                [*point, 1],
            ],
            dtype=float
        ).transpose()

        # The transform moves the transducer into the ASL skin interpolator space.
        # We want a transform that moves the transducer into LPS space, and we also want to apply the standoff transform
        transducer_transform = interpolator2lps @ transducer_transform @ standoff_transform

        # Target in transducer coordinates (lat, ele, ax)
        target_XYZ = (np.linalg.inv(transducer_transform) @ ras_lps_swap @ np.array([*target_RAS,1.0]))[:3]

        # Target in "steering space", where the origin is the center of the steering zone.
        target_steering_space = target_XYZ - np.array([0.,0.,transducer_steering_center_distance])

        steering_distance : float = float(np.linalg.norm(target_steering_space))

        # Check whether the target is in the steering range
        target_in_bounds : bool = bool(np.all((steering_mins < target_steering_space) & (target_steering_space < steering_maxs)))

        # Finally, fill out the arrays we have been building in this loop
        transducer_poses[i] = transducer_transform
        steering_dists[i] = steering_distance
        in_bounds[i] = target_in_bounds


    sorted_transforms = [
        x[0] for x in sorted(zip(transducer_poses[in_bounds],steering_dists[in_bounds]), key = lambda x : x[1])
    ]

    log.info("Virtual fitting complete.")

    return sorted_transforms



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
from __future__ import annotations

import importlib.metadata

project = 'openlifu'
copyright = '2023, Openwater'
author = 'Openwater'
version = release = importlib.metadata.version("openlifu")

import sys
from pathlib import Path

sys.path.insert(0, Path('..').resolve())

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'myst_parser',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'sphinx.ext.napoleon',
]

source_suffix = [".rst", ".md"]
templates_path = ['_templates']
exclude_patterns = [
    "_build",
    "_templates",
    "**.ipynb_checkpoints",
    "Thumbs.db",
    ".DS_Store",
    ".env",
    ".venv",
]

autosummary_generate = True  # Turn on sphinx.ext.autosummary

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = True
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'alabaster'
html_static_path = ['_static']

myst_enable_extensions = [
    "colon_fence",
]



from __future__ import annotations

import sys

from openlifu.io.LIFUInterface import LIFUInterface

# set PYTHONPATH=%cd%\src;%PYTHONPATH%
# python notebooks/test_console.py
"""
Test script to automate:
1. Connect to the device.
2. Test HVController: Turn HV on/off and check voltage.
3. Test Device functionality.
"""
print("Starting LIFU Test Script...")
interface = LIFUInterface(TX_test_mode=False)
tx_connected, hv_connected = interface.is_device_connected()
if tx_connected and hv_connected:
    print("LIFU Device Fully connected.")
else:
    print(f'LIFU Device NOT Fully Connected. TX: {tx_connected}, HV: {hv_connected}')

if not hv_connected:
       print("HV Controller not connected.")
       sys.exit()

print("Ping the device")
interface.hvcontroller.ping()

# Set High Voltage Level
if not interface.hvcontroller.set_voltage(voltage=30.0):
    print("Failed to set voltage.")

# Get Set High Voltage Setting
print("Get Current HV Voltage")
read_voltage = interface.hvcontroller.get_voltage()
print(f"HV Voltage {read_voltage} V.")


print("Test HV Supply...")
if interface.hvcontroller.turn_hv_on():
# Get Set High Voltage Setting
    read_voltage = interface.hvcontroller.get_voltage()
    print(f"HV Voltage {read_voltage} V.")
    print("HV ON Press enter to TURN OFF:")
    input()  # Wait for the user to press Enter
    if interface.hvcontroller.turn_hv_off():
        print("HV OFF.")
    else:
        print("Failed to turn off HV")
else:
    print("Failed to turn on HV.")

print("Reset DevConsoleice:")
# Ask the user for confirmation
user_input = input("Do you want to reset the Console? (y/n): ").strip().lower()

if user_input == 'y':
    if interface.hvcontroller.soft_reset():
        print("Reset Successful.")
elif user_input == 'n':
    print("Reset canceled.")
else:
    print("Invalid input. Please enter 'y' or 'n'.")



from __future__ import annotations

import asyncio
import logging
from enum import Enum
from typing import Dict

from openlifu.io.LIFUHVController import HVController
from openlifu.io.LIFUSignal import LIFUSignal
from openlifu.io.LIFUTXDevice import TxDevice
from openlifu.io.LIFUUart import LIFUUart
from openlifu.plan.solution import Solution


class LIFUInterfaceStatus(Enum):
    STATUS_COMMS_ERROR = -1
    STATUS_SYS_OFF = 0
    STATUS_SYS_POWERUP = 1
    STATUS_SYS_ON = 2
    STATUS_PROGRAMMING = 3
    STATUS_READY = 4
    STATUS_NOT_READY = 5
    STATUS_RUNNING = 6
    STATUS_FINISHED = 7
    STATUS_ERROR = 8

logger = logging.getLogger(__name__)

class LIFUInterface:
    signal_connect: LIFUSignal = LIFUSignal()
    signal_disconnect: LIFUSignal = LIFUSignal()
    signal_data_received: LIFUSignal = LIFUSignal()
    hvcontroller: HVController = None
    txdevice: TxDevice = None

    def __init__(self, vid: int = 0x0483, tx_pid: int = 0x57AF, con_pid: int = 0x57A0, baudrate: int = 921600, timeout: int = 10, TX_test_mode: bool = False, HV_test_mode: bool = False, run_async: bool = False) -> None:
        """
        Initialize the LIFUInterface with given parameters and store them in the class.

        Args:
            vid (int): Vendor ID of the USB device.
            tx_pid (int): Product ID for TX device.
            con_pid (int): Product ID for console device.
            baudrate (int): Communication baud rate.
            timeout (int): Read timeout in seconds.
            test_mode (bool): Enable test mode.
            run_async (bool): Enable asynchronous operation.
        """
        # Store parameters in instance variables
        self.vid = vid
        self.tx_pid = tx_pid
        self.con_pid = con_pid
        self.baudrate = baudrate
        self.timeout = timeout
        self._test_mode = TX_test_mode
        self._async_mode = run_async
        self._tx_uart = None
        self._hv_uart = None

        # Create a TXDevice instance as part of the interface
        logger.debug("Initializing TX Module of LIFUInterface with VID: %s, PID: %s, baudrate: %s, timeout: %s", vid, tx_pid, baudrate, timeout)
        self._tx_uart = LIFUUart(vid=vid, pid=tx_pid, baudrate=baudrate, timeout=timeout, desc="TX", demo_mode=TX_test_mode, async_mode=run_async)
        self.txdevice = TxDevice(uart=self._tx_uart)

        # Create a LIFUHVController instance as part of the interface
        logger.debug("Initializing Console of LIFUInterface with VID: %s, PID: %s, baudrate: %s, timeout: %s", vid, con_pid, baudrate, timeout)
        self._hv_uart = LIFUUart(vid=vid, pid=con_pid, baudrate=baudrate, timeout=timeout, desc="HV", demo_mode=HV_test_mode, async_mode=run_async)
        self.hvcontroller = HVController(uart=self._hv_uart)

        # Connect signals to internal handlers
        if self._async_mode:
            self._tx_uart.signal_connect.connect(self.signal_connect.emit)
            self._tx_uart.signal_disconnect.connect(self.signal_disconnect.emit)
            self._tx_uart.signal_data_received.connect(self.signal_data_received.emit)
            self._hv_uart.signal_connect.connect(self.signal_connect.emit)
            self._hv_uart.signal_disconnect.connect(self.signal_disconnect.emit)
            self._hv_uart.signal_data_received.connect(self.signal_data_received.emit)

    async def start_monitoring(self, interval: int = 1) -> None:
        """Start monitoring for USB device connections."""
        try:
            await asyncio.gather(
                self._tx_uart.monitor_usb_status(interval),
                self._hv_uart.monitor_usb_status(interval)
            )

        except Exception as e:
            logger.error("Error starting monitoring: %s", e)
            raise e

    def stop_monitoring(self) -> None:
        """Stop monitoring for USB device connections."""
        try:
            self._tx_uart.stop_monitoring()
            self._hv_uart.stop_monitoring()
        except Exception as e:
            logger.error("Error stopping monitoring: %s", e)
            raise e

    def is_device_connected(self) -> tuple:
        """
        Check if the device is currently connected.

        Returns:
            tuple: (tx_connected, hv_connected)
        """
        tx_connected = self.txdevice.is_connected()
        hv_connected = self.hvcontroller.is_connected()
        return tx_connected, hv_connected

    def set_solution(self,
                     solution: Solution | Dict,
                     profile_index:int=1,
                     profile_increment:bool=True) -> bool:
        """
        Load a solution to the device.

        Args:
            solution (Solution): The solution to load.
            profile_index (int): The profile index to load the solution to (defaults to 0)
            profile_increment (bool): Increment the profile index
        """
        try:
            #if self._test_mode:
            #    return True

            if isinstance(solution, Solution):
                solution = solution.to_dict()

            if "name" in solution:
                solution_name = solution["name"]
                solution_name = f'Solution "{solution_name}"'
            else:
                solution_name = "Solution"

            voltage = solution['pulse']['amplitude']

            logger.info("Loading %s...", solution_name)
            # Convert solution data and send to the device
            self.txdevice.set_solution(
                    pulse = solution['pulse'],
                    delays = solution['delays'],
                    apodizations= solution['apodizations'],
                    sequence= solution['sequence'],
                    profile_index=profile_index,
                    profile_increment=profile_increment
                )
            self.hvcontroller.set_voltage(voltage)
            logger.info("%s loaded successfully.", solution_name)
            return True

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle
        except Exception as e:
            logger.error("Error loading %s: %s", solution_name, e)
            raise

    def start_sonication(self) -> bool:
        """
        Start sonication.

        Sets the device to a running state and sends a start command if necessary.
        """
        try:
            if self._test_mode:
                return True

            logger.info("Turn ON HV")
            bHvOn = self.hvcontroller.turn_hv_on()

            logger.info("Start Sonication")
            # Send the solution data to the device
            bTriggerOn = self.txdevice.start_trigger()

            if bTriggerOn and bHvOn:
                logger.info("Sonication started successfully.")
                return True
            else:
                logger.error("Failed to start sonication.")
                return False

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Error Starting sonication: %s", e)
            raise e

    def get_status(self):
        """
        Query the device status.

        Returns:
            int: The device status.
        """
        if self._test_mode:
            return LIFUInterfaceStatus.STATUS_READY

        status = LIFUInterfaceStatus.STATUS_ERROR
        return status

    def stop_sonication(self) -> bool:
        """
        Stop sonication.

        Stops the current sonication process.
        """
        try:
            if self._test_mode:
                return True

            logger.info("Stop Sonication")
            # Send the solution data to the device
            bTriggerOff = self.txdevice.stop_trigger()
            bHvOff = self.hvcontroller.turn_hv_off()

            if bTriggerOff and bHvOff:
                logger.info("Sonication stopped successfully.")
                return True
            else:
                logger.error("Failed to stop sonication.")
                return False

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Error Stopping sonication: %s", e)
            raise e

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop_monitoring()
        if self.txdevice:
            self.txdevice.disconnect()
        if self.hvcontroller:
            self.hvcontroller.disconnect()



from __future__ import annotations

import numpy as np
from xarray import Dataset


def getunittype(unit):
    unit = unit.lower()
    if unit in ['micron', 'microns']:
        return 'distance'
    elif unit in ['minute', 'minutes', 'min', 'mins', 'hour', 'hours', 'hr', 'hrs', 'day', 'days', 'd']:
        return 'time'
    elif unit in ['rad', 'deg', 'radian', 'radians', 'degree', 'degrees', '°']:
        return 'angle'
    elif 'sec' in unit:
        return 'time'
    elif 'meter' in unit or 'micron' in unit:
        return 'distance'
    elif unit.endswith('s'):
        return 'time'
    elif unit.endswith('m'):
        return 'distance'
    elif unit.endswith(('m2', 'm^2')):
        return 'area'
    elif unit.endswith(('m3', 'm^3')):
        return 'volume'
    elif unit.endswith('hz'):
        return 'frequency'
    elif unit.endswith('pa'):
        return 'pascal'
    elif unit.endswith('w'):
        return 'watt'
    else:
        return 'other'

def getunitconversion(from_unit, to_unit, unitratio=None, constant=None):
    if not from_unit:
        return 1.0

    if unitratio is not None and constant is not None:
        if '/' not in unitratio:
            raise ValueError('Conversion unit ratio must have a \'/\' symbol')

        unitn, unitd = unitratio.split('/')
        type0 = getunittype(from_unit)
        type1 = getunittype(to_unit)
        typen = getunittype(unitn)
        typed = getunittype(unitd)

        if type0 == typed and type1 == typen:
            scl = getunitconversion(from_unit, unitd) * constant * getunitconversion(unitn, to_unit)
        elif type0 == typen and type1 == typed:
            scl = getunitconversion(from_unit, unitn) * 1 / constant * getunitconversion(unitd, to_unit)
        elif type0 == type1:
            scl = getunitconversion(from_unit, to_unit)
        else:
            raise ValueError(f'Unit type mismatch {type0} -> ({typen}/{typed}) -> {type1}')
    else:
        slash0 = from_unit.find('/')
        slash1 = to_unit.find('/')

        if slash0 != -1 and slash1 != -1:
            num0 = from_unit[:slash0]
            denom0 = from_unit[slash0+1:]
            num1 = to_unit[:slash1]
            denom1 = to_unit[slash1+1:]
            scl = getunitconversion(num0, num1) / getunitconversion(denom0, denom1)
        elif slash0 == -1 and slash1 == -1:
            type0 = getunittype(from_unit)
            type1 = getunittype(to_unit)

            if type0 != type1:
                raise ValueError(f'Unit type mismatch ({type0}) vs ({type1})')

            if type0 == 'other':
                if from_unit[-1] != to_unit[-1]:
                    raise ValueError(f'Cannot convert {from_unit} to {to_unit}')

                i = 0
                while i < min(len(from_unit), len(to_unit)) and from_unit[-i:] == to_unit[-i:]:
                    type = from_unit[-i:]
                    i += 1

                scl0 = getsiscale(from_unit, type)
                scl1 = getsiscale(to_unit, type)
                scl = scl0 / scl1
            else:
                scl0 = getsiscale(from_unit, type0)
                scl1 = getsiscale(to_unit, type0)
                scl = scl0 / scl1
        else:
            raise ValueError(f'Unit ratio mismatch ({from_unit} vs {to_unit})')

    return scl

def getsiscale(unit, type):
    type = type.lower()

    if type in ['distance', 'area', 'volume']:
        idx = unit.find('meters')
        if idx == -1:
            idx = unit.find('meter')
            if idx == -1:
                if unit.lower() == 'micron':
                    idx = 6
                else:
                    idx = unit.rfind('m')
                    if idx == -1:
                        idx = len(unit)

    elif type == 'time':
        idx = unit.find('seconds')
        if idx == -1:
            idx = unit.find('second')
            if idx == -1:
                idx = unit.find('sec')
                if idx == -1:
                    idx = unit.rfind('s')
                    if idx == -1:
                        idx = len(unit)

    elif type == 'angle':
        idx = len(unit)

    elif type == 'frequency' or type == "pascal":
        idx = len(unit) - 2

    elif type == "watt":
        idx = len(unit) - 1

    else:
        idx = len(unit) - len(type) + 1

    prefix = unit[:idx]

    if not prefix:
        scl = 1.0
    else:
        scl = 1.0

        if prefix == 'pico' or prefix == 'p':
            scl = 1.0e-12
        elif prefix == 'nano' or prefix == 'n':
            scl = 1.0e-9
        elif prefix == 'micro' or prefix == 'u' or prefix == '\u00b5' or prefix == '\u03bc':
            scl = 1.0e-6
        elif prefix == 'milli' or prefix == 'm':
            scl = 1.0e-3
        elif prefix == 'centi' or prefix == 'c':
            scl = 1.0e-2
        elif prefix == '':
            scl = 1.0
        elif prefix == 'kilo' or prefix == 'k':
            scl = 1.0e3
        elif prefix == 'mega' or prefix == 'M':
            scl = 1.0e6
        elif prefix == 'giga' or prefix == 'G':
            scl = 1.0e9
        elif prefix == 'tera' or prefix == 'T':
            scl = 1.0e12
        elif prefix == 'min' or prefix == 'minute':
            scl = 60.0
        elif prefix == 'hour' or prefix == 'hr':
            scl = 60.0 * 60.0
        elif prefix == 'day' or prefix == 'd':
            scl = 60.0 * 60.0 * 24.0
        elif prefix == 'rad' or prefix == 'radian' or prefix == 'radians':
            scl = 1.0
        elif prefix == 'deg' or prefix == 'degree' or prefix == 'degrees' or prefix == '\u00b0':
            scl = 2 * 3.14159265358979323846 / 360
        elif prefix:
            raise ValueError(f'Unknown prefix {prefix}')

    if type == 'area':
        scl = scl ** 2.0
    elif type == 'volume':
        scl = scl ** 3.0

    return scl


def rescale_data_arr(data_arr: Dataset, units: str) -> Dataset:
    """
    Rescales the Dataset to the specified units.

    Args:
        data_arr : xarray.Dataset
        units: str

    Returns:
        rescaled: The rescaled xarray to new units.
    """
    rescaled = data_arr.copy(deep=True)
    scale = getunitconversion(data_arr.attrs['units'], units)
    rescaled.data *= scale
    rescaled.attrs['units'] = units

    return rescaled


def rescale_coords(data_arr: Dataset, units: str) -> Dataset:
    """
    Rescales the Dataset coordinates to the specified units.

    Args:
        data_arr : xarray.Dataset
        units: str

    Returns:
        rescaled: The rescaled data_arr coords to new units.
    """
    rescaled = data_arr.copy(deep=True)
    for coord_key in data_arr.coords:
        curr_coord_attrs = rescaled[coord_key].attrs
        if 'units' in curr_coord_attrs:
            curr_coord_units = curr_coord_attrs['units']
            scale = getunitconversion(curr_coord_units, units)
            curr_coord_rescaled = scale*rescaled[coord_key].data
            rescaled = rescaled.assign_coords({coord_key: (coord_key, curr_coord_rescaled, curr_coord_attrs)})
            rescaled[coord_key].attrs['units'] = units

    return rescaled


def get_ndgrid_from_arr(data_arr: Dataset) -> np.ndarray:
    """
    Creates a ndgrid from xarray.Dataset coordinates.

    Args:
        coords : xarray.Coordinates

    Returns:
        ndgrid: The ndgrid from the Coordinates.
    """
    # First need to get correct coordinates for the ndgrid
    first_data_key = next(iter(data_arr.keys()))
    ordered_key = data_arr[first_data_key].dims
    all_coord = []
    for coord_key in ordered_key:
        if 'units' in data_arr[coord_key].attrs:
            all_coord += [data_arr.coords[coord_key].data]
    ndgrid = np.stack(np.meshgrid(*all_coord, indexing="ij"), axis=-1)

    return ndgrid



from __future__ import annotations

from . import seg_method
from .seg_method import SegmentationMethod, UniformSegmentation
from .tissue import Tissue
from .water import Water

__all__ = [
    "seg_method",
    "SegmentationMethod",
    "UniformSegmentation",
    "Water",
    "Tissue",
]



from __future__ import annotations

import sys

from openlifu.io.LIFUInterface import LIFUInterface

# set PYTHONPATH=%cd%\src;%PYTHONPATH%
# python notebooks/test_toggle_12v.py

print("Starting LIFU Test Script...")
interface = LIFUInterface(TX_test_mode=False)

tx_connected, hv_connected = interface.is_device_connected()
if tx_connected and hv_connected:
    print("LIFU Device Fully connected.")
else:
    print(f'LIFU Device NOT Fully Connected. TX: {tx_connected}, HV: {hv_connected}')

if not hv_connected:
    print("HV Controller not connected.")
    sys.exit()

print("Ping the device")
interface.hvcontroller.ping()

interface.hvcontroller.turn_12v_on()

print("12v ON. Press enter to TURN OFF:")
input()  # Wait for user input

interface.hvcontroller.turn_12v_off()



from __future__ import annotations

import sys
from asyncio import sleep

import numpy as np

from openlifu.bf.pulse import Pulse
from openlifu.bf.sequence import Sequence
from openlifu.geo import Point
from openlifu.io.LIFUInterface import LIFUInterface
from openlifu.plan.solution import Solution

# set PYTHONPATH=%cd%\src;%PYTHONPATH%
# python notebooks/test_updated_if.py
"""
Test script to automate:
1. Connect to the device.
2. Test HVController: Turn HV on/off and check voltage.
3. Test Device functionality.
"""
print("Starting LIFU Test Script...")
interface = LIFUInterface()
tx_connected, hv_connected = interface.is_device_connected()
if tx_connected and hv_connected:
    print("LIFU Device Fully connected.")
else:
    print(f'LIFU Device NOT Fully Connected. TX: {tx_connected}, HV: {hv_connected}')

if hv_connected:
    print("Ping Console device")
    interface.hvcontroller.ping()

    print("Turn 12V ON")
    interface.hvcontroller.turn_12v_on()
    sleep(2)
else:
    print("HV Controller not connected.")
    sys.exit()


if interface.txdevice.is_connected():
    print("LIFU Transmitter device connected.")
    print("Ping Transmitter device")
    interface.txdevice.ping()
else:
    print("Transmitter device not connected.")
    sys.exit()

print("Set HV to 20V")
interface.hvcontroller.set_voltage(20.0)

# Get Set High Voltage Setting
print("Get Current HV Voltage")
read_voltage = interface.hvcontroller.get_voltage()
print(f"HV Voltage {read_voltage} V.")

print("Ping Transmitter device")
interface.txdevice.ping()

print("Toggle LED")
interface.txdevice.toggle_led()

print("Get Version")
version = interface.txdevice.get_version()
print(f"Version: {version}")

print("Echo Data")
echo, echo_len = interface.txdevice.echo(echo_data=b'Hello LIFU!')
if echo_len > 0:
    print(f"Echo: {echo.decode('utf-8')}")  # Echo: Hello LIFU!
else:
    print("Echo failed.")

print("Get HW ID")
hw_id = interface.txdevice.get_hardware_id()
print(f"HWID: {hw_id}")

print("Get Temperature")
temperature = interface.txdevice.get_temperature()
print(f"Temperature: {temperature} °C")

print("Get Ambient")
a_temp = interface.txdevice.get_ambient_temperature()
print(f"Ambient Temperature: {a_temp} °C")

print("Enumerate TX7332 chips")
num_tx_devices = interface.txdevice.enum_tx7332_devices()
if num_tx_devices > 0:
    print(f"Number of TX7332 devices found: {num_tx_devices}")
else:
    raise Exception("No TX7332 devices found.")

# set focus
xInput = 0
yInput = 0
zInput = 50

frequency = 400e3
voltage = 12.0
duration = 2e-5

pulse = Pulse(frequency=frequency, amplitude=voltage, duration=duration)
pt = Point(position=(xInput,yInput,zInput), units="mm")
sequence = Sequence(
    pulse_interval=0.1,
    pulse_count=10,
    pulse_train_interval=1,
    pulse_train_count=1
)

# Calculate delays and apodizations to perform beam forming

solution = Solution(
    delays = np.zeros((1,64)),
    apodizations = np.ones((1,64)),
    pulse = pulse,
    sequence = sequence
)

sol_dict = solution.to_dict()
profile_index = 1
profile_increment = True
print("Set Solution")
interface.txdevice.set_solution(
    pulse = sol_dict['pulse'],
    delays = sol_dict['delays'],
    apodizations= sol_dict['apodizations'],
    sequence= sol_dict['sequence'],
    mode = "continuous",
    profile_index=profile_index,
    profile_increment=profile_increment
)

print("Get Trigger")
trigger_setting = interface.txdevice.get_trigger_json()
if trigger_setting:
    print(f"Trigger Setting: {trigger_setting}")
else:
    print("Failed to get trigger setting.")



print("Turn HV ON")
interface.hvcontroller.turn_hv_on()

print("Starting Trigger...")
if interface.txdevice.start_trigger():
    print("Trigger Running Press enter to STOP:")
    input()  # Wait for the user to press Enter
    if interface.txdevice.stop_trigger():
        print("Trigger stopped successfully.")
    else:
        print("Failed to stop trigger.")
else:
    print("Failed to get trigger setting.")

print("Turn HV OFF")
interface.hvcontroller.turn_hv_off()

print("Reset Device:")
# Ask the user for confirmation
user_input = input("Do you want to reset the device? (y/n): ").strip().lower()

if user_input == 'y':
    if interface.txdevice.soft_reset():
        print("Reset Successful.")
elif user_input == 'n':
    print("Reset canceled.")
else:
    print("Invalid input. Please enter 'y' or 'n'.")



from __future__ import annotations

import sys

from openlifu.io.LIFUInterface import LIFUInterface

# set PYTHONPATH=%cd%\src;%PYTHONPATH%
# python notebooks/test_pwr.py

print("Starting LIFU Test Script...")
interface = LIFUInterface(TX_test_mode=False)

tx_connected, hv_connected = interface.is_device_connected()
if tx_connected and hv_connected:
    print("LIFU Device Fully connected.")
else:
    print(f'LIFU Device NOT Fully Connected. TX: {tx_connected}, HV: {hv_connected}')

if not hv_connected:
    print("HV Controller not connected.")
    sys.exit()

print("Ping the device")
interface.hvcontroller.ping()

print("Starting DAC value increments...")
hvp_value = 2400
hrp_value = 2095
hrm_value = 2095
hvm_value = 2400

print(f"Setting DACs: hvp={hvp_value}, hrp={hrp_value}, hvm={hvm_value}, hrm={hrm_value}")
if interface.hvcontroller.set_dacs(hvp=hvp_value, hrp=hrp_value, hvm=hvm_value, hrm=hrm_value):
    print(f"DACs successfully set to hvp={hvp_value}, hrp={hrp_value}, hvm={hvm_value}, hrm={hrm_value}")
else:
    print("Failed to set DACs.")

print("Test HV Supply...")
print("HV OFF. Press enter to TURN ON:")
input()  # Wait for user input
if interface.hvcontroller.turn_hv_on():
    print("HV ON. Press enter to TURN OFF:")
    input()  # Wait for user input
    if interface.hvcontroller.turn_hv_off():
        print("HV OFF.")
    else:
        print("Failed to turn off HV.")
else:
    print("Failed to turn on HV.")



from __future__ import annotations

import threading
import time

import numpy as np

from openlifu.bf.pulse import Pulse
from openlifu.bf.sequence import Sequence
from openlifu.geo import Point
from openlifu.io.LIFUInterface import LIFUInterface
from openlifu.plan.solution import Solution
from openlifu.xdc import Transducer

# set PYTHONPATH=%cd%\src;%PYTHONPATH%
# python notebooks/test_watertank.py

"""
Test script to automate:
1. Connect to the device.
2. Test HVController: Turn HV on/off and check voltage.
3. Test Device functionality.
"""

log_interval = 1  # seconds; you can adjust this variable as needed
stop_logging = False  # flag to signal the logging thread to stop

def log_temperature():
    # Create a file with the current timestamp in the name
    timestamp = time.strftime("%Y%m%d_%H%M%S")
    filename = f"{timestamp}_temp.log"
    with open(filename, "w") as logfile:
        while not stop_logging:
            temperature = interface.txdevice.get_temperature()
            a_temp = interface.txdevice.get_ambient_temperature()
            current_time = time.strftime("%Y-%m-%d %H:%M:%S")
            log_line = f"{current_time}: Temperature: {temperature}, Ambient Temperature: {a_temp}\n"
            logfile.write(log_line)
            logfile.flush()  # Ensure the data is written immediately
            time.sleep(log_interval)


print("Starting LIFU Test Script...")
interface = LIFUInterface()
tx_connected, hv_connected = interface.is_device_connected()
if tx_connected and hv_connected:
    print("LIFU Device Fully connected.")
else:
    print(f'LIFU Device NOT Fully Connected. TX: {tx_connected}, HV: {hv_connected}')

# Ask the user if they want to log temperature
log_choice = input("Do you want to log temperature before starting trigger? (y/n): ").strip().lower()
log_temp = (log_choice == "y")

print("Ping the device")
interface.txdevice.ping()

print("Enumerate TX7332 chips")
num_tx_devices = interface.txdevice.enum_tx7332_devices()
if num_tx_devices > 0:
    print(f"Number of TX7332 devices found: {num_tx_devices}")
else:
    raise Exception("No TX7332 devices found.")

# set focus
xInput = 0
yInput = 0
zInput = 50

frequency = 405e3
voltage = 12.0
duration = 2e-5

pulse = Pulse(frequency=frequency, amplitude=voltage, duration=duration)
pt = Point(position=(xInput,yInput,zInput), units="mm")

#arr = Transducer.from_file(r"C:\Users\Neuromod2\Documents\OpenLIFU-python\OpenLIFU_2x.json")
# arr = Transducer.from_file(R"..\M4_flex.json")
arr = Transducer.from_file(R"E:\CURRENT-WORK\openwater\OpenLIFU-python\notebooks\pinmap.json")

focus = pt.get_position(units="mm")
#arr.elements = np.array(arr.elements)[np.argsort([el.pin for el in arr.elements])].tolist()
distances = np.sqrt(np.sum((focus - arr.get_positions(units="mm"))**2, 1))
tof = distances*1e-3 / 1500
delays = tof.max() - tof
apodizations = np.ones(arr.numelements())



    # tURN only single element ON
#active_element = 25

#delays = delays*0.0
#apodizations = np.zeros(arr.numelements())
#apodizations[active_element-1] = 1
print('apodizations', apodizations)
print('Delays', delays)

sequence = Sequence(
    pulse_interval=0.1,
    pulse_count=10,
    pulse_train_interval=1,
    pulse_train_count=1
)

solution = Solution(
    delays = delays,
    apodizations = apodizations,
    pulse = pulse,
    sequence = sequence
)

sol_dict = solution.to_dict()
profile_index = 1
profile_increment = True
interface.txdevice.set_solution(
    pulse = sol_dict['pulse'],
    delays = sol_dict['delays'],
    apodizations= sol_dict['apodizations'],
    sequence= sol_dict['sequence'],
    mode="continuous",
    profile_index=profile_index,
    profile_increment=profile_increment
)

print("Get Trigger")
trigger_setting = interface.txdevice.get_trigger_json()
if trigger_setting:
    print(f"Trigger Setting: {trigger_setting}")
else:
    print("Failed to get trigger setting.")

# If logging is enabled, start the logging thread
if log_temp:
    t = threading.Thread(target=log_temperature)
    t.start()
else:
    print("Get Temperature")
    temperature = interface.txdevice.get_temperature()
    print(f"Temperature: {temperature} °C")

    print("Get Ambient")
    a_temp = interface.txdevice.get_ambient_temperature()
    print(f"Ambient Temperature: {a_temp} °C")

print("Press enter to START trigger:")
input()  # Wait for the user to press Enter
print("Starting Trigger...")
if interface.txdevice.start_trigger():
    print("Trigger Running Press enter to STOP:")
    input()  # Wait for the user to press Enter
    if interface.txdevice.stop_trigger():
        stop_logging = True
        print("Trigger stopped successfully.")
    else:
        print("Failed to stop trigger.")
else:
    print("Failed to get trigger setting.")

# Stop the temperature logging before starting the trigger
if log_temp:
    t.join()



from __future__ import annotations

import re
from typing import Literal

Cases = Literal['lower', 'upper', 'same', 'snake', 'camel', 'pascal', 'cobra', 'title', 'sentence']

def sanitize(in_str, case: Cases ='snake'):
    stripped = ''.join(re.findall(r'[\w\s]+', in_str))
    words = re.findall(r'\w+', stripped)

    if case.lower() == 'lower':
        out_str = ''.join(words).lower()
    elif case.lower() == 'upper':
        out_str = ''.join(words).upper()
    elif case.lower() == 'same':
        out_str = ''.join(words)
    elif case.lower() == 'snake':
        out_str = '_'.join(words).lower()
    elif case.lower() == 'camel':
        out_str = ''.join([word[0].upper() + word[1:].lower() for word in words])
        out_str = out_str[0].lower() + out_str[1:]
    elif case.lower() == 'pascal':
        out_str = ''.join([word[0].upper() + word[1:].lower() for word in words])
    elif case.lower() == 'cobra':
        out_str = '_'.join([word[0].upper() + word[1:].lower() for word in words])
    elif case.lower() == 'title':
        words = re.findall(r'\w+', stripped.replace("_", " "))
        words = [word[0].upper() + word[1:].lower() for word in words]
        lowercase_words = ["a","an","and","for","in","of","the"]
        words = [word.lower() if word in lowercase_words else word for word in words]
        out_str = ' '.join(words)
    elif case.lower() == 'sentence':
        words = re.findall(r'\w+', stripped.replace("_", " "))
        first_word = words[0]
        words[0] = first_word[0].upper() + first_word[1:].lower()
        out_str = ' '.join(words)
    else:
        raise ValueError(f'Unrecognized case type {case}')

    return out_str



from __future__ import annotations

from openlifu.virtual_fit import VirtualFitOptions


def test_unit_conversion():
    vfo = VirtualFitOptions(
        units="cm",
        transducer_steering_center_distance=45.,
        pitch_range=(-12,14),
        yaw_range=(-13,15),
        pitch_step = 9,
        yaw_step = 10,
        planefit_dyaw_extent = 2.3,
        steering_limits=((-10,11),(-12,13),(-14,15)),
    )
    vfo_converted = vfo.to_units("mm")
    assert vfo_converted.transducer_steering_center_distance == 10*vfo.transducer_steering_center_distance
    assert vfo_converted.planefit_dyaw_extent == 10*vfo.planefit_dyaw_extent
    assert vfo_converted.yaw_step == vfo.yaw_step
    assert vfo_converted.pitch_range == vfo.pitch_range
    assert isinstance(vfo_converted.steering_limits, tuple)
    assert all(isinstance(sl, tuple) for sl in vfo_converted.steering_limits)
    assert vfo_converted.steering_limits[2][0] == 10*vfo.steering_limits[2][0]



from __future__ import annotations

from dataclasses import dataclass
from typing import Annotated

import numpy as np
import xarray as xa

from openlifu.bf.apod_methods import ApodizationMethod
from openlifu.geo import Point
from openlifu.util.annotations import OpenLIFUFieldData
from openlifu.xdc import Transducer


@dataclass
class PiecewiseLinear(ApodizationMethod):
    zero_angle: Annotated[float, OpenLIFUFieldData("Zero Apodization Angle", "Angle at and beyond which the piecewise linear apodization is 0%")] = 90.0
    """Angle at and beyond which the piecewise linear apodization is 0%"""

    rolloff_angle: Annotated[float, OpenLIFUFieldData("Rolloff start angle", "Angle below which the piecewise linear apodization is 100%")] = 45.0
    """Angle below which the piecewise linear apodization is 100%"""

    units: Annotated[str, OpenLIFUFieldData("Angle units", "Angle units")] = "deg"
    """Angle units"""

    def calc_apodization(self, arr: Transducer, target: Point, params: xa.Dataset, transform:np.ndarray | None=None):
        target_pos = target.get_position(units="m")
        matrix = transform if transform is not None else np.eye(4)
        angles = np.array([el.angle_to_point(target_pos, units="m", matrix=matrix, return_as=self.units) for el in arr.elements])
        apod = np.zeros(arr.numelements())
        f = ((self.zero_angle - angles) / (self.zero_angle - self.rolloff_angle))
        apod = np.maximum(0, np.minimum(1, f))
        return apod



from __future__ import annotations

import importlib
import json
import logging
import shutil
import tempfile
import threading
from dataclasses import dataclass
from pathlib import Path
from subprocess import PIPE, CalledProcessError, CompletedProcess, Popen
from typing import Annotated, Any, Callable, Dict, List, Tuple

import cv2
import numpy as np
import onnxruntime as ort
import OpenEXR
import requests
import trimesh
import vtk
from PIL import Image
from vtk.util.numpy_support import numpy_to_vtk

from openlifu.util.annotations import OpenLIFUFieldData

logger_meshrecon = logging.getLogger("MeshRecon")
logger_meshroom = logging.getLogger("Meshroom")

@dataclass
class Photoscan:
    id: Annotated[str, OpenLIFUFieldData("Photoscan ID", "ID of this photoscan")] = "photoscan"
    """ID of this photoscan"""

    name: Annotated[str, OpenLIFUFieldData("Photoscan name", "Photoscan name")] = "Photoscan"
    """Photoscan name"""

    model_filename: Annotated[str | None, OpenLIFUFieldData("Model filename", "Relative path to model")] = None
    """Relative path to model"""

    texture_filename: Annotated[str | None, OpenLIFUFieldData("Texture filename", "Relative path to texture image")] = None
    """Relative path to texture image"""

    mtl_filename: Annotated[str | None, OpenLIFUFieldData("Material filename", "Relative path to materials file")] = None
    """Relative path to materials file"""

    photoscan_approved: Annotated[bool, OpenLIFUFieldData("Approved?", "Approval state of the photoscan. 'True' means the user has provided some kind of confirmation that the photoscan is good enough to be used.")] = False
    """Approval state of the photoscan. 'True' means the user has provided some kind of
    confirmation that the photoscan is good enough to be used."""

    @staticmethod
    def from_json(json_string: str) -> Photoscan:
        """Load a Photoscan from a json string"""
        return Photoscan.from_dict(json.loads(json_string))

    def to_json(self, compact: bool) -> str:
        """Serialize a Photoscan to a json string. This is different to the format written to file
        and does not contain the loaded models and texture.
        Args:
            compact:if enabled then the string is compact (not pretty). Disable for pretty.
        Returns: A json string representing the complete Photoscan object
        """
        from openlifu.util.json import PYFUSEncoder

        if compact:
            return json.dumps(self.to_dict(), separators=(',', ':'), cls=PYFUSEncoder)
        else:
            return json.dumps(self.to_dict(), indent=4, cls=PYFUSEncoder)

    @staticmethod
    def from_dict(d:Dict) -> Photoscan:
        """
        Create a Photoscan from a dictionary
        param d: Dictionary of photoscan parameters.
        returns: Photoscan object
        """
        return Photoscan(**d)

    def to_dict(self) -> Dict:
        """
        Convert the photoscan to a dictionary
        returns: Dictionary of photoscan parameters
        """
        d = self.__dict__.copy()
        return d

    @staticmethod
    def from_file(filename) -> Photoscan:
        """
        Load a Photoscan from a metadata file
        :param filename: Name of the file
        """
        with open(filename) as f:
            return Photoscan.from_dict(json.load(f))

    def to_file(self, filename):
        """
        Save the photoscan to a file.
        :param filename: Name of the file
        """
        Path(filename).parent.parent.mkdir(exist_ok=True) #photoscan directory
        Path(filename).parent.mkdir(exist_ok=True)
        with open(filename, 'w') as file:
            file.write(self.to_json(compact = False))

def load_data_from_filepaths(model_abspath: str, texture_abspath: str) -> Tuple[vtk.vtkPolyData, vtk.vtkImageData]:
    """
    This function returns the data directly from the model and texture filepaths without requiring a photoscan object.
    param model_abspath: absolute filepath to model data
          texture abspath: absolute filepath to texture data
    Returns: Photoscan data as (model_vtkpolydata, texture_vtkimagedata)
    """
    model_polydata  = load_model(model_abspath)
    texture_imagedata = load_texture(texture_abspath)

    return (model_polydata, texture_imagedata)

def load_data_from_photoscan(photoscan: Photoscan, parent_dir) -> Tuple[vtk.vtkPolyData, vtk.vtkImageData]:
    """
    param parent_dir: parent directory containing model and texture data filepaths
    Returns: Photoscan data as (model_vtkpolydata, texture_vtkimagedata)
    """
    model_polydata  = load_model(Path(parent_dir)/photoscan.model_filename)
    texture_imagedata = load_texture(Path(parent_dir)/photoscan.texture_filename)

    return (model_polydata, texture_imagedata)

def load_model(file_name) -> vtk.vtkPolyData:
    """
    This function assumes that the model is saved to file in LPS coordinates.
    The model that is returned is in RAS coordinates
    """
    if not Path(file_name).exists():
        raise FileNotFoundError(f'Model filepath does not exist: {file_name}')
    mesh = read_as_vtkpolydata(file_name)
    mesh_ras = convert_between_ras_and_lps(mesh)

    return mesh_ras

def load_texture(file_name) -> vtk.vtkImageData:
    if not Path(file_name).exists():
        raise FileNotFoundError(f'Texture data filepath does not exist: {file_name}')
    texture = read_as_vtkimagedata(file_name)
    return texture

def read_as_vtkpolydata(file_name):

    suffix_to_reader_dict = {
        '.ply' : vtk.vtkPLYReader,
        '.vtp' : vtk.vtkXMLPolyDataReader,
        '.obj' : vtk.vtkOBJReader,
        '.stl' : vtk.vtkSTLReader,
        '.vtk' : vtk.vtkPolyDataReader,
        '.g' : vtk.vtkBYUReader
    }
    path = Path(file_name)
    ext = path.suffix.lower()
    if path.suffix not in suffix_to_reader_dict:
        raise ValueError(f"File format {ext} not supported by reader")
    reader = suffix_to_reader_dict[ext]()
    if ext == '.g':
        reader.SetGeometryName(file_name)
    else:
        reader.SetFileName(file_name)
    reader.Update()
    poly_data = reader.GetOutput()

    return poly_data

def convert_numpy_to_vtkimage(image_numpy: np.ndarray) -> vtk.vtkImageData:
    """
    Converts a numpy array with dimensions [HxWx3] representing an RGB image into vtkImageData
    """
    vtkimage_data = vtk.vtkImageData()
    vtkimage_data.SetDimensions(image_numpy.shape[1], image_numpy.shape[0], 1)
    vtkimage_data.SetNumberOfScalarComponents(image_numpy.shape[2], vtkimage_data.GetInformation())
    pd = vtkimage_data.GetPointData()
    new_rgb_data = image_numpy.reshape((-1, image_numpy.shape[2]))
    vtk_array = numpy_to_vtk(new_rgb_data, deep=True, array_type=vtk.VTK_UNSIGNED_SHORT)
    pd.SetScalars(vtk_array)
    return vtkimage_data

def read_as_vtkimagedata(file_name) -> vtk.vtkImageData:

    suffix_to_reader_dict = {
        '.jpg' : vtk.vtkJPEGReader,
        '.png' : vtk.vtkPNGReader,
        '.tiff' : vtk.vtkTIFFReader,
    }
    path = Path(file_name)
    ext = path.suffix.lower()
    if ext == '.exr':
        with OpenEXR.File(str(file_name), separate_channels = True) as exr_file:
            R = exr_file.channels()['R'].pixels
            G = exr_file.channels()['G'].pixels
            B = exr_file.channels()['B'].pixels
            # Combine channels into a single RGB image (H x W x 3)
            rgb_data = np.stack([R, G, B], axis=-1, dtype = np.float32)
            # EXR stores pixel data using half 16-bit floating point values. Convert data to an 8-bit image compatibility with VTK
            rgb_data = np.clip(rgb_data*(2**16-1), 0, 65535)
            image_data = convert_numpy_to_vtkimage(rgb_data)
    elif ext in suffix_to_reader_dict:
        reader = suffix_to_reader_dict[ext]()
        reader.SetFileName(file_name)
        reader.Update()
        image_data = reader.GetOutput()
    else:
        raise ValueError(f"File format {path.suffix} not supported by reader")

    return image_data

def convert_between_ras_and_lps(mesh : vtk.vtkPointSet) -> vtk.vtkPointSet:
    """Converts a mesh (polydata, unstructured grid or even just a point cloud) between the LPS (left-posterior-superior) coordinate system and
      RAS (right-anterior-superior) coordinate system."""

    transform_ras_to_lps = vtk.vtkTransform()
    transform_ras_to_lps.Scale(-1,-1,1)

    transformFilter = vtk.vtkTransformPolyDataFilter()
    transformFilter.SetInputData(mesh)
    transformFilter.SetTransform(transform_ras_to_lps)
    transformFilter.Update()

    return transformFilter.GetOutput()

def get_meshroom_pipeline_names() -> list[str]:
    """Get a list of names of valid meshroom pipelines that can be used in run_reconstruction"""
    pipeline_dir = importlib.resources.files("openlifu.nav.meshroom_pipelines")
    return [f.stem for f in pipeline_dir.iterdir() if f.suffix == ".mg"]

def subprocess_stream_output(
    args: str | List[str],
    stdout_handler: Callable[[str], None],
    stderr_handler: Callable[[str], None],
    check: bool = True,
    text: bool = True,
    **kwargs: Any,
) -> CompletedProcess:
    """
    Run a subprocess and stream its stdout and stderr output to separate handlers/loggers. A drop in replacement
    for subprocess.run that handles logging.

    Args:
        args (Union[str, List[str]]): Command and arguments to execute. Can be a string or list of strings.
        stdout_handler (Callable[[str], None]): Function to handle each line of standard output.
        stderr_handler (Callable[[str], None]): Function to handle each line of standard error.
        check (bool, optional): If True, raise CalledProcessError if the subprocess exits with a non-zero code.
        text (bool, optional): If True, communicate with the process using text mode. Defaults to True.
        **kwargs (Any): Additional keyword arguments passed to subprocess.Popen.

    Returns:
        CompletedProcess: An object containing the arguments used and the return code.

    Raises:
        CalledProcessError: If `check` is True and the subprocess exits with a non-zero status.
    """
    with Popen(args, stdout=PIPE, stderr=PIPE, text=text, **kwargs) as process:

        def log_stream(stream, handler):
            for line in stream:
                handler(line.rstrip())

        # Create threads for each stream
        stdout_thread = threading.Thread(target=log_stream, args=(process.stdout, stdout_handler))
        stderr_thread = threading.Thread(target=log_stream, args=(process.stderr, stderr_handler))
        stdout_thread.start()
        stderr_thread.start()

        stdout_thread.join()
        stderr_thread.join()

        retcode = process.wait()

    if check and retcode:
        raise CalledProcessError(retcode, process.args)
    return CompletedProcess(process.args, retcode)

def run_reconstruction(images: list[Path],
                       pipeline_name: str = "default_pipeline",
                       input_resize_width: int = 3024,
                       use_masks: bool = True) -> Tuple[Photoscan,Path]:
    """Run Meshroom with the given images and pipeline.
    Args:
        images (list[Path]): List of image file paths.
        pipeline_name (str): Name of the Meshroom pipeline in meshroom_pipelines folder.
            See also `get_meshroom_pipeline_names`.
        input_resize_width (int): Width to which input images will be resized, in pixels.
        use_masks (bool): Whether to include a background removal step to filter the dense reconstruction.

    Returns:
        photoscan: The Photoscan of the reconstructed images.
        data_dir (Path): The directory containing the underlying data files whose names are given in the Photoscan.
    """
    pipeline_dir = importlib.resources.files("openlifu.nav.meshroom_pipelines")
    valid_configs = get_meshroom_pipeline_names()

    if pipeline_name not in valid_configs:
        raise ValueError(
            f"Invalid pipeline name '{pipeline_name}'. "
            f"Valid options are: {', '.join(valid_configs)}"
        )

    pipeline = pipeline_dir / f"{pipeline_name}.mg"

    if shutil.which("meshroom_batch") is None:
        raise FileNotFoundError("Error: 'meshroom_batch' is not found in system PATH. Ensure it is installed and accessible.")

    temp_dir = Path(tempfile.mkdtemp())

    images_dir = temp_dir / "images"
    images_dir.mkdir(parents=True, exist_ok=True)

    #resize the images and store in tmp dir
    new_paths = []
    for image in images:
        img = Image.open(image)
        orientation = img.getexif().get(274, 1)
        if orientation in range(4,8):
            #camera is rotated. Get the perceived width.
            old_width = img.height
        else:
            old_width = img.width
        scale = input_resize_width/old_width

        resize_width = int(scale*img.width)
        resize_height = int(scale*img.height)

        img = img.resize((resize_width, resize_height), Image.Resampling.LANCZOS)
        img.save( images_dir / image.name, exif=img.getexif())
        new_paths.append(images_dir / image.name)

    output_dir = temp_dir / "output"
    cache_dir = temp_dir / "cache"

    command = [
        "meshroom_batch",
        "--pipeline", pipeline.as_posix(),
        "--output", output_dir.as_posix(),
        "--input", images_dir.as_posix(),
        "--cache", cache_dir.as_posix()
    ]

    if use_masks:
        masks_dir = temp_dir / "masks"
        masks_dir.mkdir(parents=True, exist_ok=True)
        make_masks(new_paths, masks_dir)
        command += ["--paramOverrides", f"PrepareDenseScene_1.masksFolders=['{masks_dir.as_posix()}']"]

    subprocess_stream_output(command, logger_meshroom.info, logger_meshroom.error)

    output_dir_merged = temp_dir / "output_dir_merged"
    output_dir_merged.mkdir(parents=True, exist_ok=True)

    merge_textures(output_dir / "texturedMesh.obj", output_dir_merged)

    photoscan_dict = {
        "model_filename":  "texturedMesh.obj",
        "texture_filename": "material_0.png",
        "mtl_filename": "material.mtl",
    }

    photoscan = Photoscan.from_dict(photoscan_dict)

    return photoscan, output_dir_merged

def udim_to_tile(udim_str: str) -> Tuple[int, int]:
    """Convert UDIM string to tile coordinates"""
    x = int(udim_str[-2:]) - 1
    tile_u = x % 10
    tile_v = x // 10
    return tile_u, tile_v

def merge_textures(input_obj_path: Path, output_path: Path) -> None:
    """Merge the UDIM textures output by meshroom into a single large texture

    Args:
        input_obj_path (Path): Path to .obj mesh to merge.
        output_path (Path): Path to save merged .obj mesh.
    """
    scene = trimesh.load(input_obj_path, process=True)

    if isinstance(scene, trimesh.Scene):
        mesh_dict = scene.geometry
    else:
        mesh_dict = {'material_1001': scene}

    for name, mesh in mesh_dict.items():
        mesh.tile = udim_to_tile(name[-4:])

    num_u_tiles = max(mesh.tile[0] for _, mesh in mesh_dict.items()) + 1
    num_v_tiles = max(mesh.tile[1] for _, mesh in mesh_dict.items()) + 1

    first_tile = np.array(mesh_dict['material_1001'].visual.material.image)
    tile_height, tile_width, tile_ch = first_tile.shape
    tex_height, tex_width = num_v_tiles*tile_height, num_u_tiles*tile_height
    new_texture = np.zeros((tex_height, tex_width, tile_ch), dtype=first_tile.dtype)

    new_verts = []
    new_faces = []
    new_uvs = []
    num_verts = 0
    for _, mesh in mesh_dict.items():
        tile_u, tile_v = mesh.tile
        texture_tile = np.array(mesh.visual.material.image)
        new_texture[tex_height - tile_height*(tile_v+1): tex_height-tile_height*tile_v, tile_width*tile_u:tile_width*(tile_u+1)] = texture_tile
        new_verts.append(mesh.vertices)
        new_faces.append(mesh.faces+num_verts)
        new_uvs.append(mesh.visual.uv)
        num_verts += mesh.vertices.shape[0]

    new_texture = Image.fromarray(new_texture)
    new_verts = np.concatenate(new_verts, axis=0)
    new_faces = np.concatenate(new_faces, axis=0)
    new_uvs = np.concatenate(new_uvs, axis=0)

    new_uvs = new_uvs/np.array([num_u_tiles, num_v_tiles]).reshape(-1,2)

    mesh = trimesh.Trimesh() #Note putting vertices in this constructor will merge nearby ones
    mesh.vertices = new_verts
    mesh.faces = new_faces
    mesh.visual = trimesh.visual.TextureVisuals(uv=new_uvs, image=new_texture)

    mesh.export(output_path / "texturedMesh.obj")


def apply_exif_orientation_numpy(image: np.ndarray, orientation: int, inverse: bool = False) -> np.ndarray:
    """
    Transforms an image array to undo or redo EXIF orientation.

    Args:
        image (np.ndarray): The image array (H x W x C or H x W).
        orientation (int): EXIF orientation tag value (1-8).
        inverse (bool): If True, applies the inverse transformation.

    Returns:
        np.ndarray: The transformed image.
    """
    if orientation <= 1 or orientation > 8:
        return image  # No transformation needed

    # Define forward transformations
    def transform(img):
        if orientation == 2:
            return np.fliplr(img)
        elif orientation == 3:
            return np.rot90(img, 2)
        elif orientation == 4:
            return np.flipud(img)
        elif orientation == 5:
            return np.rot90(np.fliplr(img), 1)
        elif orientation == 6:
            return np.rot90(img, -1)
        elif orientation == 7:
            return np.rot90(np.fliplr(img), -1)
        elif orientation == 8:
            return np.rot90(img, 1)
        else:
            raise ValueError("Unknown exif orientation")

    # Define inverse transformations
    def inverse_transform(img):
        if orientation == 2:
            return np.fliplr(img)
        elif orientation == 3:
            return np.rot90(img, 2)
        elif orientation == 4:
            return np.flipud(img)
        elif orientation == 5:
            return np.fliplr(np.rot90(img, -1))
        elif orientation == 6:
            return np.rot90(img, 1)
        elif orientation == 7:
            return np.fliplr(np.rot90(img, 1))
        elif orientation == 8:
            return np.rot90(img, -1)
        else:
            raise ValueError("Unknown exif orientation")

    return inverse_transform(image) if inverse else transform(image)


def get_modnet_path() -> Path:
    """Get the MODNet checkpoint path. Download it if not present.
    """
    package = "openlifu.nav.modnet_checkpoints"
    filename = "modnet_photographic_portrait_matting.onnx"
    url = "https://data.kitware.com/api/v1/file/67feb2cb31a330568827ab32/download"
    try:
        # Try to find the checkpoint in the package
        resource_path = importlib.resources.files(package) / filename
        if resource_path.is_file():
            logger_meshrecon.info(f"Found existing MODNet checkpoint at {resource_path}")
            return resource_path
    except (FileNotFoundError, ModuleNotFoundError):
        pass

    # Fallback: Download the checkpoint
    base_dir = Path(importlib.resources.files(package))
    full_path = base_dir / filename
    logger_meshrecon.info(f"MODNet checkpoint not found. Downloading from {url}...")
    response = requests.get(url, stream=True, timeout=(10, 300))
    if response.status_code == 200:
        with open(full_path, 'wb') as f:
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:
                    f.write(chunk)
        logger_meshrecon.info(f"Downloaded MODNet checkpoint to {full_path}")
    else:
        raise RuntimeError(f"Failed to download MODNet checkpoint: {response.status_code} - {response.text}")

    return full_path

def preprocess_image_modnet(image: np.ndarray, ref_size: int = 512) -> np.ndarray:
    """
    Preprocess an input image for MODNet inference.

    This function performs the same preprocessing steps as the official MODNet code:
    - Normalizes image values to the range [-1, 1]
    - Resizes the image based on a reference size (512), maintaining aspect ratio
    - Ensures the resized dimensions are divisible by 32
    - Converts the image to CHW format and adds a batch dimension

    Args:
        image (np.ndarray): Input image in HWC format with values in [0, 255].

    Returns:
        np.ndarray: Preprocessed image in NCHW format with float32 values in [-1, 1].
    """

    # Normalize image to [-1, 1]
    image = image.astype(np.float32) / 255.0
    image = 2 * image - 1

    im_h, im_w, _ = image.shape

    # Resize if both dimensions are smaller or larger than the reference size
    if max(im_h, im_w) < ref_size or min(im_h, im_w) > ref_size:
        if im_w >= im_h:
            im_rh = ref_size
            im_rw = int(im_w * ref_size / im_h)
        else:
            im_rw = ref_size
            im_rh = int(im_h * ref_size / im_w)
    else:
        im_rh, im_rw = im_h, im_w

    # Ensure dimensions are divisible by 32
    im_rw -= im_rw % 32
    im_rh -= im_rh % 32

    # Resize, convert to CHW, and add batch dimension
    image = cv2.resize(image, (im_rw, im_rh), interpolation=cv2.INTER_AREA)
    image = np.transpose(image, (2, 0, 1))  # HWC to CHW
    image = np.expand_dims(image, axis=0)  # Add batch dimension

    return image


def make_masks(image_paths: list[Path], output_dir: Path, threshold: float = 0.01) -> None:
    """
    Runs MODNet on a list of image paths and saves the output masks.

    Each output mask is saved in `output_dir` using the original filename with a `.png` extension.
    The `threshold` parameter is used to convert the MODNet soft segmentation output into a binary (hard) mask.
    EXIF orientation data is preserved to ensure correct image alignment when loading into Meshroom.

    Args:
        image_paths (List[str]): List of input image file paths.
        output_dir (str): Directory where the output masks will be saved.
        threshold (float): Threshold to binarize the soft segmentation output.
    """
    # Load the ONNX model
    ckpt_path = get_modnet_path()
    session = ort.InferenceSession(ckpt_path, providers=["CPUExecutionProvider"])  # or CUDAExecutionProvider
    for image_path in image_paths:
        image = Image.open(image_path)
        exif = image.getexif()
        orientation = exif.get(274,1)
        image = np.array(image)
        image = apply_exif_orientation_numpy(image, orientation=orientation)
        im_h, im_w, _ = image.shape
        image = preprocess_image_modnet(image)

        # Run the model
        mask = session.run(["output"], {"input": image})[0].squeeze()
        mask = cv2.resize(mask, (im_w, im_h), interpolation=cv2.INTER_AREA)
        mask = mask > threshold
        mask = (mask.astype(np.float32)*255).astype(np.uint8)
        mask = np.repeat(mask[:, :, np.newaxis], 3, axis=2)

        mask = apply_exif_orientation_numpy(mask, orientation=orientation, inverse=True)

        mask = Image.fromarray(mask)
        output_path = output_dir / (image_path.stem + ".png")
        mask.save(output_path, exif=exif)



from __future__ import annotations

import copy
from abc import abstractmethod
from dataclasses import asdict, dataclass, field
from typing import Annotated

import numpy as np
import xarray as xa

from openlifu.seg import seg_methods
from openlifu.seg.material import MATERIALS, PARAM_INFO, Material
from openlifu.util.annotations import OpenLIFUFieldData


@dataclass
class SegmentationMethod:
    materials: Annotated[dict[str, Material], OpenLIFUFieldData("Segmentation materials", "Dictionary mapping of label names to material definitions used during segmentation")] = field(default_factory=lambda: MATERIALS.copy())
    """Dictionary mapping of label names to material definitions used during segmentation"""

    ref_material: Annotated[str, OpenLIFUFieldData("Reference material", "Reference material ID to use")] = "water"
    """Reference material ID to use"""

    def __post_init__(self):
        if self.ref_material not in self.materials:
            raise ValueError(f"Reference material {self.ref_material} not found.")
    @abstractmethod
    def _segment(self, volume: xa.DataArray):
        pass

    @staticmethod
    def from_dict(d):
        if isinstance(d, str):
            if d == "water":
                return seg_methods.Water()
            elif d == "tissue":
                return seg_methods.Tissue()
            elif d == "segmented":
                return seg_methods.SegmentMRI()
        else:
            d = copy.deepcopy(d)
            short_classname = d.pop("class")
            if "materials" in d:
                for material_key, material_definition in d["materials"].items():
                    if not isinstance(material_definition, Material): # if it is given as a dict rather than a fully hydrated object
                        d["materials"][material_key] = Material.from_dict(material_definition)
            module_dict = seg_methods.__dict__
            class_constructor = module_dict[short_classname]
            return class_constructor(**d)

    def _material_indices(self, materials: dict | None = None):
        materials = self.materials if materials is None else materials
        return {material_id: i for i, material_id in enumerate(materials.keys())}

    def _map_params(self, seg: xa.DataArray, materials: dict | None = None):
        materials = self.materials if materials is None else materials
        material_dict = self._material_indices(materials=materials)
        params = xa.Dataset()
        ref_mat = materials[self.ref_material]
        for param_id in PARAM_INFO:
            info = Material.param_info(param_id)
            param = xa.DataArray(np.zeros(seg.shape), coords=seg.coords, attrs={"units": info["units"], "long_name": info["name"], "ref_value": ref_mat.get_param(param_id)})
            for material_id, material in materials.items():
                midx = material_dict[material_id]
                param.data[seg.data == midx] = getattr(material, param_id)
            params[param_id] = param
        params.attrs['ref_material'] = ref_mat
        return params

    def seg_params(self, volume: xa.DataArray, materials: dict | None = None):
        materials = self.materials if materials is None else materials
        seg = self._segment(volume)
        params = self._map_params(seg, materials=materials)
        return params

    def ref_params(self, coords: xa.Coordinates):
        seg = self._ref_segment(coords)
        params = self._map_params(seg)
        return params

    def _ref_segment(self, coords: xa.Coordinates):
        material_dict = self._material_indices()
        m_idx = material_dict[self.ref_material]
        sz = list(coords.sizes.values())
        seg = xa.DataArray(np.full(sz, m_idx, dtype=int), coords=coords)
        return seg

    def to_dict(self):
        d = asdict(self)
        d['class'] = self.__class__.__name__
        return d

@dataclass
class UniformSegmentation(SegmentationMethod):
    def _segment(self, vol: xa.DataArray):
        return self._ref_segment(vol.coords)



from __future__ import annotations

import copy
import json
from dataclasses import dataclass, field
from typing import Annotated, Any, Dict, Tuple

import numpy as np
import vtk

from openlifu.util.annotations import OpenLIFUFieldData
from openlifu.util.dict_conversion import DictMixin
from openlifu.util.units import getunitconversion


# === Tools to work with points ===
@dataclass
class Point:
    position: Annotated[np.ndarray, OpenLIFUFieldData("Position", "3D position of the point in the provided units")] = field(default_factory=lambda: np.array([0.0, 0.0, 0.0]))  # mm
    """3D position of the point in the provided units"""

    id: Annotated[str, OpenLIFUFieldData("Point ID", "Unique identifier for the point")] = "point"
    """Unique identifier for the point"""

    name: Annotated[str, OpenLIFUFieldData("Point name", "Name of the point")] = "Point"
    """Name of the point"""

    color: Annotated[Any, OpenLIFUFieldData("Color (RGB)", "RGB color of the point")] = (1.0, 0.0, 0.0)
    """RGB color of the point"""

    radius: Annotated[float, OpenLIFUFieldData("Radius", "Radius for rendering the point in the provided units")] = 1.0  # mm
    """Radius for rendering the point in the provided units"""

    dims: Annotated[Tuple[str, str, str], OpenLIFUFieldData("Dimensions", "Names of the axes of the coordinate system being used")] = ("x", "y", "z")
    """Names of the axes of the coordinate system being used"""

    units: Annotated[str, OpenLIFUFieldData("Units", "Units for the point")] = "mm"
    """Units for the point"""

    def __post_init__(self):
        if len(self.position) != len(self.dims):
            raise ValueError("Position and dims must have same length.")
        self.position = np.array(self.position).reshape(3)

    def copy(self):
        return copy.deepcopy(self)

    def get_position(self, dim=None, units: str | None =None):
        units = self.units if units is None else units
        scl = getunitconversion(self.units, units)
        if dim is None:
            return self.position*scl
        else:
            return self.position[self.dims.index(dim)]*scl

    def get_matrix(self, origin: np.ndarray = np.eye(4), center_on_point: bool = True, local: bool = False):
        pos = np.dot(np.linalg.inv(origin), np.append(self.position, 1.0))[:3]
        if center_on_point:
            center = pos
        else:
            center = np.zeros(3)
        zvec = np.array([0., 0., 1.])
        if np.linalg.norm(pos) != 0:
            zvec = pos / np.linalg.norm(pos)
        az = -np.arctan2(zvec[0], zvec[2])
        xvec = np.array([np.cos(az), 0.0, np.sin(az)])
        yvec = np.cross(zvec, xvec)
        m = np.array([[xvec[0], yvec[0], zvec[0], center[0]],
                      [xvec[1], yvec[1], zvec[1], center[1]],
                      [xvec[2], yvec[2], zvec[2], center[2]],
                      [0.0, 0.0, 0.0, 1.0]])
        if not local:
            m = np.dot(origin, m)
        return m

    def get_polydata(self, transform: np.ndarray = np.eye(4), units=None):
        units = self.units if units is None else units
        colors =  vtk.vtkNamedColors()
        # Create a sphere
        sphereSource = vtk.vtkSphereSource()
        scl = getunitconversion(self.units, units)
        pos = np.dot(transform, np.append(self.position*scl, 1.0))[:3]
        sphereSource.SetCenter(*pos)
        sphereSource.SetRadius(self.radius*scl)
        sphereSource.SetPhiResolution(100)
        sphereSource.SetThetaResolution(100)
        return sphereSource

    def get_actor(self, transform: np.ndarray = np.eye(4), units=None):
        polydata = self.get_polydata(transform=transform, units=units)
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(polydata.GetOutputPort())
        actor = vtk.vtkActor()
        actor.SetMapper(mapper)
        actor.GetProperty().SetColor(self.color)
        return actor

    def rescale(self, units: str):
        scl = getunitconversion(self.units, units)
        self.position = self.position * scl
        self.radius = self.radius * scl
        self.units = units

    def transform(self,
                  matrix: np.ndarray,
                  units: str | None = None,
                  new_dims: Tuple[str, str, str] | None=None):
        if units is not None:
            self.rescale(units)
        self.position = np.dot(matrix, np.append(self.position, 1.0))[:3]
        if new_dims is not None:
            self.dims = new_dims

    def to_dict(self):
        return {"id": self.id,
                "name": self.name,
                "color": self.color,
                "radius": self.radius,
                "position": self.position.tolist(),
                "dims": self.dims,
                "units": self.units}

    @staticmethod
    def from_dict(point_data:Dict):
        """Create a Point object from a dictionary."""
        if "color" in point_data:
            if len(point_data["color"]) != 3:
                raise ValueError(f"Color should have three components; got {point_data['color']}.")
            point_data["color"] = tuple(float(point_data["color"][i]) for i in range(3))
        if "radius" in point_data:
            point_data["radius"] = float(point_data["radius"])
        if "position" in point_data:
            point_data["position"] = np.array(point_data["position"])
        if "dims" in point_data:
            point_data["dims"] = tuple(point_data["dims"])
        return Point(**point_data)

    @staticmethod
    def from_json(json_string : str) -> Point:
        """Load a Point from a json string"""
        return Point.from_dict(json.loads(json_string))

    def to_json(self, compact:bool) -> str:
        """Serialize a Point to a json string

        Args:
            compact: if enabled then the string is compact (not pretty). Disable for pretty.

        Returns: A json string representing the complete Point object.
        """
        if compact:
            return json.dumps(self.to_dict(), separators=(',', ':'))
        else:
            return json.dumps(self.to_dict(), indent=4)


@dataclass
class ArrayTransform(DictMixin):
    """An affine transform with a unit string, often intended to represent how a transducer array is positioned in space."""

    matrix: Annotated[np.ndarray, OpenLIFUFieldData("Affine matrix", "4x4 affine transform matrix")]
    """4x4 affine transform matrix"""

    units: Annotated[str, OpenLIFUFieldData("Units", "The units of the space on which to apply the transform matrix , e.g. 'mm' (In order to apply the transform to points, first represent the points in these units.)")]
    """The units of the space on which to apply the transform matrix , e.g. "mm"
    (In order to apply the transform to points, first represent the points in these units.)
    """

# === Tools to work with spherical coordinate systems ===

def cartesian_to_spherical(x:float,y:float,z:float) -> Tuple[float, float, float]:
    """Convert cartesian coordinates to spherical coordinates

    Args: x, y, z are cartesian coordinates
    Returns: r, theta, phi, where
        r is the radial spherical coordinate, a nonnegative float.
        theta is the polar spherical coordinate, aka the angle off the z-axis, aka the non-azimuthal spherical angle.
            theta is in the range [0,pi].
        phi is the azimuthal spherical coordinate, in the range [-pi,pi]

    Angles are in radians.
    """
    return (np.sqrt(x**2+y**2+z**2), np.arctan2(np.sqrt(x**2+y**2),z), np.arctan2(y,x))

def spherical_to_cartesian(r: float, th:float, ph:float) -> Tuple[float, float, float]:
    """Convert spherical coordinates to cartesian coordinates

    Args:
        r: the radial spherical coordinate
        th: the polar spherical coordinate theta, aka the angle off the z-axis, aka the non-azimuthal spherical angle
        ph: the azimuthal spherical coordinate phi
    Returns the cartesian coordinates x,y,z

    Angles are in radians.
    """
    return (r*np.sin(th)*np.cos(ph), r*np.sin(th)*np.sin(ph), r*np.cos(th))

def cartesian_to_spherical_vectorized(p:np.ndarray) -> np.ndarray:
    """Convert cartesian coordinates to spherical coordinates

    Args:
        p: an array of shape  (...,3), where the last axis describes point cartesian coordinates x,y,z.
    Returns: An array of shape (...,3), where the last axis describes point spherical coordinates r, theta, phi, where
        r is the radial spherical coordinate, a nonnegative float.
        theta is the polar spherical coordinate, aka the angle off the z-axis, aka the non-azimuthal spherical angle.
            theta is in the range [0,pi].
        phi is the azimuthal spherical coordinate, in the range [-pi,pi]

    Angles are in radians.
    """
    return np.stack([
        np.sqrt((p**2).sum(axis=-1)),
        np.arctan2(np.sqrt((p[...,0:2]**2).sum(axis=-1)),p[...,2]),
        np.arctan2(p[...,1],p[...,0]),
    ], axis=-1)

def spherical_to_cartesian_vectorized(p:np.ndarray) -> np.ndarray:
    """Convert spherical coordinates to cartesian coordinates

    Args:
        p: an array of shape  (...,3), where the last axis describes point spherical coordinates r, theta, phi, where:
            r is the radial spherical coordinate
            theta is the polar spherical coordinate, aka the angle off the z-axis, aka the non-azimuthal spherical angle
            phi is the azimuthal spherical coordinate
    Returns the cartesian coordinates x,y,z

    Angles are in radians.
    """
    return np.stack([
        p[...,0]*np.sin(p[...,1])*np.cos(p[...,2]),
        p[...,0]*np.sin(p[...,1])*np.sin(p[...,2]),
        p[...,0]*np.cos(p[...,1]),
    ], axis=-1)

def spherical_coordinate_basis(th:float, phi:float) -> np.ndarray:
    """Return normalized spherical coordinate basis at a location with spherical polar and azimuthal coordinates (th, phi).
    The coordinate basis is returned as an array `basis` of shape (3,3), where the rows are the basis vectors,
    in the order r, theta, phi. So `basis[0], basis[1], basis[2]` are the vectors $\\hat{r}$, $\\hat{\\theta}$, $\\hat{\\phi}$.
    Angles are assumed to be provided in radians."""
    return np.array([
        [np.sin(th)*np.cos(phi), np.sin(th)*np.sin(phi), np.cos(th)],
        [np.cos(th)*np.cos(phi), np.cos(th)*np.sin(phi), -np.sin(th)],
        [-np.sin(phi), np.cos(phi), 0],
    ])

# === General coordinate transformation utiltiies ===

def create_standoff_transform(z_offset:float, dzdy:float) -> np.ndarray:
    """Create a standoff transform based on a z_offset and a dzdy value.

    A "standoff transform" applies a displacement in transducer space that moves a transducer to where it would
    be situated with the standoff in place. The idea is that if you start with a transform that places a transducer
    directly against skin, then pre-composing that transform by a "standoff transform" serves to nudge the transducer
    such that there is space for the standoff to be between it and the skin.

    This function assumes that the standoff is laterally symmetric, has some thickness, and can raise the bottom of
    the transducer a bit more than the top. The `z_offset` is the thickness in the middle of the standoff,
    while the `dzdy` is the elevational slope.

    Args:
        z_offset: Thickness in the middle of the standoff
        dzdy: Slope of the standoff, as axial displacement per unit elevational displacement. A positive number
            here means that the bottom of the transducer is raised a little bit more than the top.

    Returns a 4x4 matrix representing a rigid transform in whatever units z_offset was provided in.
    """
    angle = np.arctan(dzdy)
    return np.array([
        [1,0,0,0],
        [0,np.cos(angle),-np.sin(angle),0],
        [0,np.sin(angle),np.cos(angle),-z_offset],
        [0,0,0,1],
    ], dtype=float)
